/*
 * Copyright 2004-2005 Ivan Leo Puoti
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "winreg.h"
#include "winternl.h"
#include "wine/debug.h"
#include "wine/unicode.h"
#include "winnt.h"
#include "ntstatus.h"
#include "ntoskrnl.h"
WINE_DEFAULT_DEBUG_CHANNEL(ntoskrnl);

KSYSTEM_TIME KeTickCount = { 0x12345678, 3, 3 };

extern DRIVERDATABASE *DriverDatabase;

NTSTATUS WINAPI IoCreateDevice(IN PDRIVER_OBJECT DriverObject, IN ULONG DeviceExtensionSize,
                               IN PUNICODE_STRING DeviceName, IN DEVICE_TYPE DeviceType,
                               IN ULONG DeviceCharacteristics, IN BOOLEAN Exclusive,
                               OUT PDEVICE_OBJECT *DeviceObject) 
{
    PDRIVERDATABASE LocalDB = DriverDatabase;

    WINE_TRACE("(%p, %ld, %s, %ld, %ld, %d, %p)\n", DriverObject, DeviceExtensionSize, wine_dbgstr_wn(DeviceName->Buffer,-1),
               DeviceType, DeviceCharacteristics, Exclusive, DeviceObject);

    while(LocalDB->down)
        LocalDB = LocalDB->down;
    while(LocalDB->Driver->DeviceObject)
        LocalDB->Driver->DeviceObject = LocalDB->Driver->DeviceObject->NextDevice;
    LocalDB->Driver->DeviceObject = RtlAllocateHeap(GetProcessHeap(), 0, sizeof(DEVICE_OBJECT));
    LocalDB->Driver->DeviceObject->Reserved = RtlAllocateHeap(GetProcessHeap(), 0, sizeof(DEVICE_NAME));
    RtlCreateUnicodeString(&(((DEVICE_NAME*)LocalDB->Driver->DeviceObject->Reserved)->DeviceName),
                           DeviceName->Buffer);
    if(DeviceExtensionSize > 0)
        LocalDB->Driver->DriverExtension = RtlAllocateHeap(GetProcessHeap(), 0, DeviceExtensionSize);
    LocalDB->Driver->DeviceObject->DeviceType = DeviceType;
    LocalDB->Driver->DeviceObject->Characteristics = DeviceCharacteristics;
    LocalDB->Driver->DeviceObject->DriverObject = LocalDB->Driver;
    LocalDB->Driver->DeviceObject->NextDevice = NULL;
    if(Exclusive == TRUE)
        WINE_FIXME("Exclusive == TRUE, unkown behaviour, please report\n");
    *DeviceObject = (PDEVICE_OBJECT)LocalDB->Driver->DeviceObject;
    return STATUS_SUCCESS;
}

VOID WINAPI IoDeleteDevice(IN PDEVICE_OBJECT DeviceObject)
{
    PDEVICE_OBJECT previous = NULL;
    PDRIVERDATABASE LocalDB = find_driver(NULL, NULL, DeviceObject, previous, NULL);
    if(LocalDB)
    {
        if(previous)
            previous->NextDevice = LocalDB->Driver->DeviceObject->NextDevice;
        RtlFreeHeap(GetProcessHeap(), 0, LocalDB->Driver->DeviceObject->DeviceExtension);
        RtlFreeHeap(GetProcessHeap(), 0, LocalDB->Driver->DeviceObject);
    }
}

NTSTATUS WINAPI IoCreateSymbolicLink(IN PUNICODE_STRING SymbolicLinkName, IN PUNICODE_STRING DeviceName)
{
    PDRIVERDATABASE LocalDB = DriverDatabase;
    int found = FALSE;

    WINE_TRACE("(%s, %s)\n", wine_dbgstr_wn(SymbolicLinkName->Buffer, -1), wine_dbgstr_wn(DeviceName->Buffer, -1));

    while(found == FALSE)
    {
        if(RtlEqualUnicodeString(&(((DEVICE_NAME*)LocalDB->Driver->DeviceObject->Reserved)->DeviceName), DeviceName, TRUE))
        {
            found = TRUE;
            RtlCreateUnicodeString(&(((DEVICE_NAME*)LocalDB->Driver->DeviceObject->Reserved)->DosDevicesName),
                                   SymbolicLinkName->Buffer + 12);
            break;
        }
        if(LocalDB->Driver->DeviceObject->NextDevice)
        {
            LocalDB->Driver->DeviceObject = LocalDB->Driver->DeviceObject->NextDevice;
            continue;
        }
        if(LocalDB->down)
            LocalDB = LocalDB->down;
        else
            break;
    }
    if(!found)
        return STATUS_NO_SUCH_DEVICE;
    else
        return STATUS_SUCCESS;
}

NTSTATUS WINAPI IoDeleteSymbolicLink(IN PUNICODE_STRING SymbolicLinkName)
{
    WINE_FIXME("(%s), stub!\n", wine_dbgstr_wn(SymbolicLinkName->Buffer, -1));
    return STATUS_SUCCESS;
}

/* __attribute__((fastcall)) */
VOID WINAPI IofCompleteRequest(/*IN PIRP Irp, IN CCHAR PriorityBoost*/VOID)
{
#if 0 /* uncomment before release */
    WINE_FIXME("stub!\n");
#endif
}

BOOLEAN WINAPI PsGetVersion(PULONG MajorVersion OPTIONAL, PULONG MinorVersion OPTIONAL,
                            PULONG BuildNumber OPTIONAL, PUNICODE_STRING CSDVersion OPTIONAL)
{
    static const WCHAR ver[] = {'S','e','r','v','i','c','e',' ','P','a','c','k',' ','4',0};

    WINE_TRACE("(%p, %p, %p, %p)\n", MajorVersion, MinorVersion, BuildNumber, CSDVersion);
    if (MajorVersion)
        *MajorVersion = 5;
    if (MinorVersion)
        *MinorVersion = 0;
    if (BuildNumber)
        *BuildNumber = 2195;
    if (CSDVersion)
    {
        CSDVersion->Length = 0;
        RtlAppendUnicodeToString(CSDVersion, ver);
    }
    return 0;
}
